@@ -1,20 +1,13 @@
 package Cassandra::Client::Connection;
-our $AUTHORITY = 'cpan:TVDW';
-$Cassandra::Client::Connection::VERSION = '0.14_003'; # TRIAL
-
-$Cassandra::Client::Connection::VERSION = '0.14003';use 5.010;
+use 5.008;
 use strict;
 use warnings;
 use vars qw/$BUFFER/;
 
-use Ref::Util qw/is_blessed_ref is_plain_arrayref/;
+use Ref::Util qw/is_arrayref/;
 use IO::Socket::INET;
-use IO::Socket::INET6;
-use Errno qw/EAGAIN/;
+use Errno;
 use Socket qw/SOL_SOCKET IPPROTO_TCP SO_KEEPALIVE TCP_NODELAY/;
-use Scalar::Util qw/weaken/;
-use Net::SSLeay qw/ERROR_WANT_READ ERROR_WANT_WRITE ERROR_NONE/;
-
 use Cassandra::Client::Util;
 use Cassandra::Client::Protocol qw/
     :constants
@@ -26,18 +19,22 @@ use Cassandra::Client::Protocol qw/
     pack_shortbytes
     pack_stringmap
     pack_stringlist
-    unpack_errordata
     unpack_inet
     unpack_int
     unpack_metadata
     unpack_shortbytes
     unpack_string
-    unpack_stringlist
     unpack_stringmultimap
 /;
-use Cassandra::Client::Error::Base;
+use Cassandra::Client::Encoder qw/
+    make_encoder
+/;
+use Cassandra::Client::Decoder qw/
+    make_decoder
+/;
+use Cassandra::Client::Error;
 use Cassandra::Client::ResultSet;
-use Cassandra::Client::TLSHandling;
+use Scalar::Util qw/weaken/;
 
 use constant STREAM_ID_LIMIT => 32768;
 
@@ -60,7 +57,6 @@ sub new {
         prepare_cache   => $args{metadata}->prepare_cache,
         last_stream_id  => 0,
         pending_streams => {},
-        in_prepare      => {},
 
         decompress_func => undef,
         compress_func   => undef,
@@ -71,14 +67,6 @@ sub new {
         pending_write   => undef,
         shutdown        => 0,
         read_buffer     => \(my $empty= ''),
-        bytes_sent      => 0,
-        bytes_read      => 0,
-
-        tls             => undef,
-        tls_want_write  => undef,
-
-        healthcheck     => undef,
-        protocol_version => $args{options}{protocol_version},
     }, $class;
     weaken($self->{async_io});
     weaken($self->{client});
@@ -91,7 +79,7 @@ sub get_local_status {
     series([
         sub {
             my ($next)= @_;
-            $self->execute_prepared($next, \"select key, data_center, host_id, broadcast_address, rack, release_version, tokens, schema_version from system.local");
+            $self->execute_prepared($next, \"select key, data_center, host_id, broadcast_address, rack, release_version, tokens from system.local", undef, { consistency => 'local_one' });
         },
         sub {
             my ($next, $result)= @_;
@@ -104,7 +92,6 @@ sub get_local_status {
                 rack => $_->[4],
                 release_version => $_->[5],
                 tokens => $_->[6],
-                schema_version => $_->[7],
             } } @{$result->rows};
 
             $next->(undef, \%local_status);
@@ -120,7 +107,7 @@ sub get_peers_status {
     series([
         sub {
             my ($next)= @_;
-            $self->execute_prepared($next, \"select peer, data_center, host_id, preferred_ip, rack, release_version, tokens, schema_version from system.peers");
+            $self->execute_prepared($next, \"select peer, data_center, host_id, preferred_ip, rack, release_version, tokens from system.peers", undef, { consistency => 'local_one' });
         },
         sub {
             my ($next, $result)= @_;
@@ -133,7 +120,6 @@ sub get_peers_status {
                 rack => $_->[4],
                 release_version => $_->[5],
                 tokens => $_->[6],
-                schema_version => $_->[7],
             } } @{$result->rows};
 
             $next->(undef, \%network_status);
@@ -176,20 +162,20 @@ sub register_events {
 
 ###### QUERY CODE
 sub execute_prepared {
-    my ($self, $callback, $queryref, $parameters, $attr, $exec_info)= @_;
+    my ($self, $callback, $queryref, $parameters, $attr)= @_;
 
     # Note: parameters is retained until the query is complete. It must not be changed; clone if needed.
     # Same for attr. Note that external callers automatically have their arguments cloned.
 
     my $prepared= $self->{prepare_cache}{$$queryref} or do {
-        return $self->prepare_and_try_execute_again($callback, $queryref, $parameters, $attr, $exec_info);
+        return $self->prepare_and_try_execute_again($callback, $queryref, $parameters, $attr);
     };
 
-    my $want_result_metadata= !$prepared->{decoder};
+    my $want_result_metadata= !$prepared->{result_metadata}{columns};
     my $row;
     if ($parameters) {
         eval {
-            $row= $prepared->{encoder}->encode($parameters);
+            $row= $prepared->{encoder}->($parameters);
             1;
         } or do {
             my $error= $@ || "??";
@@ -197,9 +183,10 @@ sub execute_prepared {
         };
     }
 
-    my $consistency= $consistency_lookup{$attr->{consistency} || 'one'};
+    my $chosen_consistency= $attr->{consistency} || $self->{options}{default_consistency};
+    my $consistency= ($chosen_consistency ? $consistency_lookup{$chosen_consistency} : CONSISTENCY_ONE);
     if (!defined $consistency) {
-        return $callback->("Invalid consistency level specified: $attr->{consistency}");
+        return $callback->("Invalid consistency level specified");
     }
 
     my $page_size= (0+($attr->{page_size} || $self->{options}{max_page_size} || 0)) || undef;
@@ -211,18 +198,14 @@ sub execute_prepared {
         my ($err, $code)= @_;
 
         if ($err) {
-            if (is_blessed_ref($err) && $err->code == 0x2500) {
-                return $self->prepare_and_try_execute_again($callback, $queryref, $parameters, $attr, $exec_info);
+            if (ref $err && $err->code == 0x2500) {
+                return $self->prepare_and_try_execute_again($callback, $queryref, $parameters, $attr);
             }
             return $callback->($err);
         }
 
         if ($code != OPCODE_RESULT) {
-            # This shouldn't ever happen...
-            return $callback->(Cassandra::Client::Error::Base->new(
-                message         => "Expected a RESULT frame but got something else; considering the query failed",
-                request_error   => 1,
-            ));
+            return $callback->("Expected a RESULT frame but got something else; considering the query failed");
         }
 
         $self->decode_result($callback, $prepared, $_[2]);
@@ -234,9 +217,9 @@ sub execute_prepared {
 }
 
 sub prepare_and_try_execute_again {
-    my ($self, $callback, $queryref, $parameters, $attr, $exec_info)= @_;
+    my ($self, $callback, $queryref, $parameters, $attr)= @_;
 
-    if ($exec_info->{_prepared_and_tried_again}++) {
+    if ($attr->{_prepared_and_tried_again}++) {
         return $callback->("Query failed because it seems to be missing from the server's prepared statement cache");
     }
 
@@ -253,28 +236,28 @@ sub prepare_and_try_execute_again {
             return $callback->("Internal error: expected query to be prepared but it was not");
         }
 
-        return $self->execute_prepared($callback, $queryref, $parameters, $attr, $exec_info);
+        return $self->execute_prepared($callback, $queryref, $parameters, $attr);
     });
     return;
 }
 
 sub execute_batch {
-    my ($self, $callback, $queries, $attribs, $exec_info)= @_;
+    my ($self, $callback, $queries, $attribs)= @_;
     # Like execute_prepared, assumes ownership of $queries and $attribs
 
-    if (!is_plain_arrayref($queries)) {
+    if (!is_arrayref($queries)) {
         return $callback->("execute_batch: queries argument must be an array of arrays");
     }
 
     my @prepared;
     for my $query (@$queries) {
-        if (!is_plain_arrayref($query)) {
+        if (!is_arrayref($query)) {
             return $callback->("execute_batch: entries in query argument must be arrayrefs");
         }
         if (!$query->[0]) {
             return $callback->("Empty or no query given, cannot execute as part of a batch");
         }
-        if ($query->[1] && !is_plain_arrayref($query->[1])) {
+        if ($query->[1] && !is_arrayref($query->[1])) {
             return $callback->("Query parameters to batch() must be given as an arrayref");
         }
 
@@ -282,7 +265,7 @@ sub execute_batch {
             push @prepared, [ $prep, $query->[1] ];
 
         } else {
-            return $self->prepare_and_try_batch_again($callback, $queries, $attribs, $exec_info);
+            return $self->prepare_and_try_batch_again($callback, $queries, $attribs);
         }
     }
 
@@ -294,14 +277,15 @@ sub execute_batch {
         }
     }
 
-    my $consistency= $consistency_lookup{$attribs->{consistency} || 'one'};
+    my $chosen_consistency= $attribs->{consistency} || $self->{options}{default_consistency};
+    my $consistency= ($chosen_consistency ? $consistency_lookup{$chosen_consistency} : CONSISTENCY_ONE);
     if (!defined $consistency) {
-        return $callback->("Invalid consistency level specified: $attribs->{consistency}");
+        return $callback->("Invalid consistency level specified");
     }
 
     my $batch_frame= pack('Cn', $batch_type, (0+@prepared));
     for my $prep (@prepared) {
-        $batch_frame .= pack('C', 1).pack_shortbytes($prep->[0]{id}).$prep->[0]{encoder}->encode($prep->[1]);
+        $batch_frame .= pack('C', 1).pack_shortbytes($prep->[0]{id}).$prep->[0]{encoder}->($prep->[1]);
     }
     $batch_frame .= pack('nC', $consistency, 0);
 
@@ -310,18 +294,14 @@ sub execute_batch {
         my ($err, $code)= @_;
 
         if ($err) {
-            if (is_blessed_ref($err) && $err->code == 0x2500) {
-                return $self->prepare_and_try_batch_again($callback, $queries, $attribs, $exec_info);
+            if (ref $err && $err->code == 0x2500) {
+                return $self->prepare_and_try_batch_again($callback, $queries, $attribs);
             }
             return $callback->($err);
         }
 
         if ($code != OPCODE_RESULT) {
-            # This shouldn't ever happen...
-            return $callback->(Cassandra::Client::Error::Base->new(
-                message         => "Expected a RESULT frame but got something else; considering the batch failed",
-                request_error   => 1,
-            ));
+            return $callback->("Expected a RESULT frame but got something else; considering the query failed");
         }
 
         $self->decode_result($callback, undef, $_[2]);
@@ -333,9 +313,9 @@ sub execute_batch {
 }
 
 sub prepare_and_try_batch_again {
-    my ($self, $callback, $queries, $attribs, $exec_info)= @_;
+    my ($self, $callback, $queries, $attribs)= @_;
 
-    if ($exec_info->{_prepared_and_tried_again}++) {
+    if ($attribs->{_prepared_and_tried_again}++) {
         return $callback->("Batch failed because one or more queries seem to be missing from the server's prepared statement cache");
     }
 
@@ -350,7 +330,7 @@ sub prepare_and_try_batch_again {
     ], sub {
         return $callback->($_[0]) if $_[0];
 
-        return $self->execute_batch($callback, $queries, $attribs, $exec_info);
+        return $self->execute_batch($callback, $queries, $attribs);
     });
     return;
 }
@@ -358,12 +338,7 @@ sub prepare_and_try_batch_again {
 sub prepare {
     my ($self, $callback, $query)= @_;
 
-    if (exists $self->{in_prepare}{$query}) {
-        push @{$self->{in_prepare}{$query}}, $callback;
-        return;
-    }
-
-    $self->{in_prepare}{$query}= [ $callback ];
+    # XXX Should we guard against the case of preparing the same statement in parallel?
 
     series([
         sub {
@@ -383,25 +358,23 @@ sub prepare {
             }
 
             my $id= unpack_shortbytes($body);
+            my $metadata= eval { unpack_metadata($body) } or return $next->("Unable to unpack query metadata: $@");
+            my $resultmetadata= eval { unpack_metadata($body) } or return $next->("Unable to unpack query metadata: $@");
 
             my ($encoder, $decoder);
             eval {
-                ($encoder)= unpack_metadata($self->{protocol_version}, 0, $body);
+                $encoder= make_encoder($metadata);
+                $decoder= make_decoder($resultmetadata);
                 1;
-            } or return $next->("Unable to unpack query metadata: $@");
-            eval {
-                ($decoder)= unpack_metadata($self->{protocol_version}, 1, $body);
-                1;
-            } or return $next->("Unable to unpack query result metadata: $@");
+            } or do {
+                my $error= $@ || "??";
+                return $callback->("Error while preparing query, couldn't compile encoder/decoder: $error");
+            };
 
-            $self->{metadata}->add_prepared($query, $id, $decoder, $encoder);
-            return $next->();
+            $self->{metadata}->add_prepared($query, $id, $metadata, $resultmetadata, $decoder, $encoder);
+            $next->();
         },
-    ], sub {
-        my $error= shift;
-        my $in_prepare= delete($self->{in_prepare}{$query}) or die "BUG";
-        $_->($error) for @$in_prepare;
-    });
+    ], $callback);
 
     return;
 }
@@ -411,20 +384,21 @@ sub decode_result {
 
     my $result_type= unpack('l>', substr($_[3], 0, 4, ''));
     if ($result_type == RESULT_ROWS) { # Rows
-        my ($paging_state, $decoder);
+        my $metadata= eval { unpack_metadata($_[3]) } or return $callback->("Unable to unpack query metadata: $@");
+        my $rows;
         eval {
-            ($decoder, $paging_state)= unpack_metadata($self->{protocol_version}, 1, $_[3]);
+            my $decoder= $prepared->{decoder} || make_decoder($metadata);
+            $rows= $decoder->($_[3]);
             1;
         } or do {
-            return $callback->("Unable to unpack query metadata: $@");
+            my $error= $@ || "??";
+            return $callback->("Error while decoding row: $error");
         };
-        $decoder= $prepared->{decoder} || $decoder;
-
         $callback->(undef,
             Cassandra::Client::ResultSet->new(
-                \$_[3],
-                $decoder,
-                $paging_state,
+                $rows,
+                [ map { $_->[2] } @{$prepared->{result_metadata}{columns} || $metadata->{columns}} ],
+                $metadata->{paging_state},
             )
         );
 
@@ -436,10 +410,7 @@ sub decode_result {
         return $callback->();
 
     } elsif ($result_type == RESULT_SCHEMA_CHANGE) { # Schema change
-        return $self->wait_for_schema_agreement(sub {
-            # We may be passed an error. Ignore it, our query succeeded
-            $callback->();
-        });
+        return $self->wait_for_schema_agreement($callback);
 
     } else {
         return $callback->("Query executed successfully but got an unexpected response type");
@@ -449,40 +420,8 @@ sub decode_result {
 
 sub wait_for_schema_agreement {
     my ($self, $callback)= @_;
-
-    my $waited= 0;
-    my $wait_delay= 0.5;
-    my $max_wait= 20;
-
-    my $done;
-    whilst(
-        sub { !$done },
-        sub {
-            my ($whilst_next)= @_;
-
-            $self->get_network_status(sub {
-                my ($error, $network_status)= @_;
-                return $whilst_next->($error) if $error;
-
-                my %versions;
-                $versions{$_->{schema_version}}= 1 for values %$network_status;
-                if (keys %versions > 1) {
-                    if ($waited >= $max_wait) {
-                        return $whilst_next->("wait_for_schema_agreement timed out after $waited seconds");
-                    }
-
-                    $waited += $wait_delay;
-                    return $self->{async_io}->timer($whilst_next, $wait_delay);
-                } else {
-                    $done= 1;
-                    return $whilst_next->();
-                }
-            });
-        },
-        $callback,
-    );
-
-    return;
+    sleep 1; # I am very sorry.
+    return $callback->();
 }
 
 
@@ -558,7 +497,7 @@ sub handshake {
         sub {
             my ($next)= @_;
             if ($self->{options}{keyspace}) {
-                return $self->execute_prepared($next, \('use "'.$self->{options}{keyspace}.'"'));
+                return $self->execute_prepared($next, \('use "'.$self->{options}{keyspace}.'"'), undef, { consistency => 'local_one' });
             }
             return $next->();
         },
@@ -589,10 +528,8 @@ sub handshake {
 sub authenticate {
     my ($self, $callback, $authenticator)= @_;
 
-    my $user= "$self->{options}{username}";
-    my $pass= "$self->{options}{password}";
-    utf8::encode($user) if utf8::is_utf8($user);
-    utf8::encode($pass) if utf8::is_utf8($pass);
+    my $user= $self->{options}{username};
+    my $pass= $self->{options}{password};
 
     if (!$user || !$pass) {
         return $callback->("Server expected authentication using <$authenticator> but no credentials were set");
@@ -652,46 +589,25 @@ sub connect {
     my ($self, $callback)= @_;
     return $callback->() if $self->{connected};
 
-    if ($self->{connecting}++) {
-        warn "BUG: Calling connect twice?";
-        return $callback->("Internal bug: called connect twice.");
+    if ($self->{connecting}) {
+        push @{$self->{connecting}}, $callback;
         return;
     }
 
-    if ($self->{options}{tls}) {
-        eval {
-            $self->{tls}= $self->{client}{tls}->new_conn;
-            1;
-        } or do {
-            my $error= $@ || "unknown TLS error";
-            return $callback->($error);
-        };
-    }
+    $self->{connecting}= [$callback];
 
     my $socket; {
         local $@;
 
-        if ($self->{host} =~ /:/) {
-            # IPv6
-            $socket= IO::Socket::INET6->new(
-                PeerAddr => $self->{host},
-                PeerPort => $self->{options}{port},
-                Proto    => 'tcp',
-                Blocking => 0,
-            );
-        } else {
-            # IPv6
-            $socket= IO::Socket::INET->new(
-                PeerAddr => $self->{host},
-                PeerPort => $self->{options}{port},
-                Proto    => 'tcp',
-                Blocking => 0,
-            );
-        }
+        $socket= IO::Socket::INET->new(
+            PeerAddr => $self->{host},
+            PeerPort => $self->{options}{port},
+            Proto    => 'tcp',
+            Blocking => 0,
+        );
 
         unless ($socket) {
-            my $error= "Could not connect: $@";
-            return $callback->($error);
+            return $callback->("Could not connect: $@");
         }
 
         $socket->setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1);
@@ -700,37 +616,28 @@ sub connect {
 
     $self->{socket}= $socket;
     $self->{fileno}= $socket->fileno;
-    $self->{async_io}->register($self->{fileno}, $self);
-    $self->{async_io}->register_read($self->{fileno});
-
-    # We create a fake buffer, to ensure we wait until we can actually write
-    $self->{pending_write}= '';
-    $self->{async_io}->register_write($self->{fileno});
-
-    if ($self->{options}{tls}) {
-        Net::SSLeay::set_fd(${$self->{tls}}, $self->{fileno});
-        Net::SSLeay::set_connect_state(${$self->{tls}});
-    }
-
+    $self->setup_io;
     $self->handshake(sub {
-        my $error= shift;
+        my $st= shift;
         $self->{connected}= 1;
-        if ($error) {
-            $self->shutdown("Failed to connect: $error");
-        }
-        return $callback->($error);
+        $_->($st) for @{$self->{connecting}};
+        undef $self->{connecting};
     });
 
     return;
 }
 
+sub setup_io {
+    my ($self)= @_;
+    $self->{async_io}->register($self->{fileno}, $self);
+    $self->{async_io}->register_read($self->{fileno});
+    return;
+}
+
 sub request {
     # my $body= $_[3] (let's avoid copying that blob). Yes, this code assumes ownership of the body.
     my ($self, $cb, $opcode)= @_;
-    return $cb->(Cassandra::Client::Error::Base->new(
-        message => "Connection shutting down",
-        request_error => 1,
-    )) if $self->{shutdown};
+    return $cb->("Connection shutting down") if $self->{shutdown};
 
     my $pending= $self->{pending_streams};
 
@@ -738,10 +645,7 @@ sub request {
     my $attempts= 0;
     while (exists($pending->{$stream_id}) || $stream_id >= STREAM_ID_LIMIT) {
         $stream_id= (++$stream_id) % STREAM_ID_LIMIT;
-        return $cb->(Cassandra::Client::Error::Base->new(
-            message => "Cannot find a stream ID to post query with",
-            request_error => 1,
-        )) if ++$attempts >= STREAM_ID_LIMIT;
+        return $cb->("Cannot find a stream ID to post query with") if ++$attempts >= STREAM_ID_LIMIT;
     }
     $self->{last_stream_id}= $stream_id;
     $pending->{$stream_id}= [$cb, $self->{async_io}->deadline($self->{fileno}, $stream_id, $self->{request_timeout})];
@@ -754,86 +658,25 @@ sub request {
             $compress_func->($_[3]);
         }
 
-        my $data= pack('CCsCN/a', $self->{protocol_version}, $flags, $stream_id, $opcode, $_[3]);
+        my $data= pack('CCsCN/a', 3, $flags, $stream_id, $opcode, $_[3]);
 
-        if (defined $self->{pending_write}) {
+        if (length $self->{pending_write}) {
+            # XXX Now that we have timeouts, we should consider not sending anything if we already timed out
             $self->{pending_write} .= $data;
             last WRITE;
         }
 
-        if ($self->{tls}) {
-            my $length= length $data;
-            my $rv= Net::SSLeay::write(${$self->{tls}}, $data);
-            if ($rv == $length) {
-                $self->{bytes_sent} += $rv;
-                # All good
-            } elsif ($rv > 0) {
-                # Partital write
-                substr($data, 0, $rv, '');
-                $self->{bytes_sent} += $rv;
-                $self->{pending_write}= $data;
-                $self->{async_io}->register_write($self->{fileno});
-            } else {
-                $rv= Net::SSLeay::get_error(${$self->{tls}}, $rv);
-                if ($rv == ERROR_WANT_WRITE || $rv == ERROR_WANT_READ || $rv == ERROR_NONE) {
-                    # Ok...
-                    $self->{pending_write}= $data;
-                    if ($rv == ERROR_WANT_READ) {
-                        $self->{tls_want_write}= 1;
-                    } else {
-                        $self->{async_io}->register_write($self->{fileno});
-                    }
-                } else {
-                    # We failed to send the request.
-                    my $error= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
-
-                    # We never actually sent our request, so take it out again
-                    my $my_stream= delete $pending->{$stream_id};
-
-                    # Disable our stream's deadline
-                    ${$my_stream->[1]}= 1;
-
-                    $self->shutdown($error);
-
-                    # Now fail our stream properly, but include the retry notice
-                    $my_stream->[0]->(Cassandra::Client::Error::Base->new(
-                        message       => "Disconnected: $error",
-                        do_retry      => 1,
-                        request_error => 1,
-                    ));
-                }
-            }
-
+        my $length= length $data;
+        my $result= syswrite($self->{socket}, $data, $length);
+        if ($result && $result == $length) {
+            # All good
+        } elsif (defined $result || $!{EAGAIN}) {
+            substr($data, 0, $result, '') if $result;
+            $self->{pending_write}= $data;
+            $self->{async_io}->register_write($self->{fileno});
         } else {
-            my $length= length $data;
-            my $result= syswrite($self->{socket}, $data, $length);
-            if ($result && $result == $length) {
-                $self->{bytes_sent} += $result;
-                # All good
-            } elsif (defined $result || $! == EAGAIN) {
-                substr($data, 0, $result, '') if $result;
-                $self->{bytes_sent} += $result;
-                $self->{pending_write}= $data;
-                $self->{async_io}->register_write($self->{fileno});
-            } else {
-                # Oh, we failed to send out the request. That's bad. Let's first find out what happened.
-                my $error= $!;
-
-                # We never actually sent our request, so take it out again
-                my $my_stream= delete $pending->{$stream_id};
-
-                # Disable our stream's deadline
-                ${$my_stream->[1]}= 1;
-
-                $self->shutdown($error);
-
-                # Now fail our stream properly, but include the retry notice
-                $my_stream->[0]->(Cassandra::Client::Error::Base->new(
-                    message       => "Disconnected: $error",
-                    do_retry      => 1,
-                    request_error => 1,
-                ));
-            }
+            my $error= $!;
+            $self->shutdown(undef, $error);
         }
     }
 
@@ -848,52 +691,19 @@ sub can_read {
 
 READ:
     while (!$self->{shutdown}) {
-        my $should_read_more;
+        my $read_cnt= sysread($self->{socket}, $BUFFER, 10240, $bufsize);
+        my $read_something;
+        if ($read_cnt) {
+            $bufsize += $read_cnt;
+            $read_something= 1;
 
-        if ($self->{tls}) {
-            my ($bytes, $rv)= Net::SSLeay::read(${$self->{tls}});
-            if (length $bytes) {
-                $BUFFER .= $bytes;
-                $bufsize += $rv;
-                $should_read_more= 1;
-                $self->{bytes_read} += $rv;
-            }
-
-            if ($rv <= 0) {
-                $rv= Net::SSLeay::get_error(${$self->{tls}}, $rv);
-                if ($rv == ERROR_WANT_WRITE) {
-                    $self->{async_io}->register_write($self->{fileno});
-                } elsif ($rv == ERROR_WANT_READ) {
-                    # Can do! Wait for the next event.
-
-                    # Resume our write if needed.
-                    if (delete $self->{tls_want_write}) {
-                        # Try our write again!
-                        $self->{async_io}->register_write($self->{fileno});
-                    }
-                } elsif ($rv == ERROR_NONE) {
-                    # Huh?
-                } else {
-                    my $error= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
-                    $shutdown_when_done= "TLS error: $error";
-                }
-            }
-
-        } else {
-            my $read_cnt= sysread($self->{socket}, $BUFFER, 16384, $bufsize);
-            if ($read_cnt) {
-                $bufsize += $read_cnt;
-                $should_read_more= 1 if $read_cnt >= 16384;
-                $self->{bytes_read} += $read_cnt;
-
-            } elsif (!defined $read_cnt) {
-                if ($! != EAGAIN) {
-                    my $error= "$!";
-                    $shutdown_when_done= $error;
-                }
-            } elsif ($read_cnt == 0) { # EOF
-                $shutdown_when_done= "Disconnected from server";
+        } elsif (!defined $read_cnt) {
+            if (!$!{EAGAIN}) {
+                my $error= "$!";
+                $shutdown_when_done= $error;
             }
+        } elsif ($read_cnt == 0) { # EOF
+            $shutdown_when_done= "Disconnected from server";
         }
 
 READ_NEXT:
@@ -907,33 +717,23 @@ READ_NEXT:
         my $body= substr($BUFFER, 0, $bodylen, '');
         $bufsize -= 9 + $bodylen;
 
+        # Decompress if needed
         if (($flags & 1) && $body) {
-            # Decompress if needed
             $self->{decompress_func}->($body);
         }
-        if ($flags & 4) {
-            # FIXME: If we reach this (we shouldn't!), we're corrupting the user's data.
-            warn 'BUG: unexpectedly received custom QueryHandler payload';
-        }
-        if ($flags & 8) {
-            # Warnings were sent from the server. Relay them.
-            my $warnings= unpack_stringlist($body);
-            for my $warning (@$warnings) {
-                warn $warning;
-            }
-        }
 
         if ($stream_id != -1) {
             my $stream_cb= delete $self->{pending_streams}{$stream_id};
             if (!$stream_cb) {
-                warn 'BUG: received response for unknown stream';
+                if (!$self->{shutdown}) {
+                    warn 'BUG: received response for unknown stream';
+                } # Else: totally fine
 
             } elsif ($opcode == OPCODE_ERROR) {
+                my ($code, $message)= unpack('Nn/a', $body);
                 my ($cb, $dl)= @$stream_cb;
                 $$dl= 1;
-
-                my $error= unpack_errordata($body);
-                $cb->($error);
+                $cb->(Cassandra::Client::Error->new($code, $message));
 
             } else {
                 my ($cb, $dl)= @$stream_cb;
@@ -948,11 +748,11 @@ READ_NEXT:
         goto READ_NEXT;
 
 READ_MORE:
-        last READ unless $should_read_more;
+        last READ unless $read_something;
     }
 
     if ($shutdown_when_done) {
-        $self->shutdown($shutdown_when_done);
+        $self->shutdown(undef, $shutdown_when_done);
     }
 
     return;
@@ -961,54 +761,21 @@ READ_MORE:
 sub can_write {
     my ($self)= @_;
 
-    if ($self->{tls}) {
-        my $rv= Net::SSLeay::write(${$self->{tls}}, $self->{pending_write});
-        if ($rv > 0) {
-            substr($self->{pending_write}, 0, $rv, '');
-            $self->{bytes_sent} += $rv;
-            if (!length $self->{pending_write}) {
-                $self->{async_io}->unregister_write($self->{fileno});
-                delete $self->{pending_write};
-            }
-            return;
-
-        } else {
-            $rv= Net::SSLeay::get_error(${$self->{tls}}, $rv);
-            if ($rv == ERROR_WANT_WRITE) {
-                # Wait until the next callback.
-                return;
-            } elsif ($rv == ERROR_WANT_READ) {
-                # Unschedule ourselves
-                $self->{async_io}->unregister_write($self->{fileno});
-                $self->{tls_want_write}= 1;
-                return;
-            } elsif ($rv == ERROR_NONE) {
-                # Huh?
-                return;
-            } else {
-                my $error= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
-                return $self->shutdown("TLS error: $error");
-            }
+    my $result= syswrite($self->{socket}, $self->{pending_write});
+    if (!defined($result)) {
+        if ($!{EAGAIN}) {
+            return; # Huh. Oh well, whatever
         }
 
-    } else {
-        my $result= syswrite($self->{socket}, $self->{pending_write});
-        if (!defined($result)) {
-            if ($! == EAGAIN) {
-                return; # Huh. Oh well, whatever
-            }
+        my $error= "$!";
+        undef $self->{pending_write};
+        return $self->shutdown(undef, $error);
+    }
+    if ($result == 0) { return; } # No idea whether that happens, but guard anyway.
+    substr($self->{pending_write}, 0, $result, '');
 
-            my $error= "$!";
-            return $self->shutdown($error);
-        }
-        if ($result == 0) { return; } # No idea whether that happens, but guard anyway.
-        substr($self->{pending_write}, 0, $result, '');
-        $self->{bytes_sent} += $result;
-
-        if (!length $self->{pending_write}) {
-            $self->{async_io}->unregister_write($self->{fileno});
-            delete $self->{pending_write};
-        }
+    if (!length $self->{pending_write}) {
+        $self->{async_io}->unregister_write($self->{fileno});
     }
 
     return;
@@ -1018,70 +785,33 @@ sub can_timeout {
     my ($self, $id)= @_;
     my $stream= delete $self->{pending_streams}{$id};
     $self->{pending_streams}{$id}= [ sub{}, \(my $zero= 0) ]; # fake it
-    $stream->[0]->(Cassandra::Client::Error::Base->new(
-        message         => "Request timed out",
-        is_timeout      => 1,
-        request_error   => 1,
-    ));
-    $self->maybe_healthcheck;
+    $stream->[0]->("Request timed out");
     return;
 }
 
-sub maybe_healthcheck {
-    my ($self)= @_;
-    return if $self->{healthcheck};
-    my $check= $self->{healthcheck}= {};
-
-    # Mark how many bytes we've read right now
-    $check->{bytes_read}= $self->{bytes_read};
-
-    series([
-        sub {
-            my ($next)= @_;
-            $self->request($next, OPCODE_OPTIONS, '');
-        }
-    ], sub {
-        my ($error)= @_;
-        $self->{healthcheck}= undef;
-        if ($self->{bytes_read} > $check->{bytes_read}) {
-            # Don't care what happened, the connection is still fine
-        } elsif (!$error) {
-            # All good
-        } elsif (is_blessed_ref($error) && $error->is_request_error && $error->is_timeout) {
-            # Not good.
-            $self->shutdown("health check timed out");
-        } else {
-            # No clue. :-)
-        }
-    });
-}
-
 sub shutdown {
-    my ($self, $shutdown_reason)= @_;
+    my ($self, $cb, $shutdown_reason)= @_;
+
+    if ($self->{shutdown}) {
+        $cb->() if $cb;
+        return;
+    }
 
-    return if $self->{shutdown};
     $self->{shutdown}= 1;
 
     my $pending= $self->{pending_streams};
     $self->{pending_streams}= {};
+    $_->[0]->("Disconnected: $shutdown_reason") for values %$pending;
 
-    # Disable our deadlines
-    ${$_->[1]}= 1 for values %$pending;
-
+    $self->{socket}->close;
     $self->{async_io}->unregister_read($self->{fileno});
-    if (defined(delete $self->{pending_write})) {
+    if (length(delete $self->{pending_write})) {
         $self->{async_io}->unregister_write($self->{fileno});
     }
     $self->{async_io}->unregister($self->{fileno}, $self);
     $self->{client}->_disconnected($self->get_pool_id);
-    $self->{socket}->close;
 
-    for (values %$pending) {
-        $_->[0]->(Cassandra::Client::Error::Base->new(
-            message       => "Disconnected: $shutdown_reason",
-            request_error => 1,
-        ));
-    }
+    $cb->() if $cb;
 
     return;
 }
@@ -1090,7 +820,7 @@ sub shutdown {
 
 ###### COMPRESSION
 BEGIN {
-    @compression_preference= qw/lz4 snappy/;
+    @compression_preference= qw/snappy lz4/;
 
     %available_compression= (
         snappy  => scalar eval "use Compress::Snappy (); 1;",
@@ -1145,28 +875,3 @@ sub decompress_lz4 {
 }
 
 1;
-
-__END__
-
-=pod
-
-=head1 NAME
-
-Cassandra::Client::Connection
-
-=head1 VERSION
-
-version 0.14_003
-
-=head1 AUTHOR
-
-Tom van der Woerdt <tvdw@cpan.org>
-
-=head1 COPYRIGHT AND LICENSE
-
-This software is copyright (c) 2019 by Tom van der Woerdt.
-
-This is free software; you can redistribute it and/or modify it under
-the same terms as the Perl 5 programming language system itself.
-
-=cut
